"""Extra widgets for Qt."""

from loguru import logger

__version__ = "0.1.4"
__author__ = "Lukasz G. Migas"
__email__ = "lukas.migas@yahoo.com"
__issue_url__ = "https://github.com/lukasz-migas/qtextra-issues/issues"
__project_url__ = "https://lukasz-migas.com/qtextra"
logger.disable("qtextra")
