"""Init."""

from qtextra.config.events import EVENTS
from qtextra.config.theme import CANVAS, THEMES, CanvasThemes, Themes, is_dark


def get_settings():
    """Get settings."""
    pass


__all__ = ["CANVAS", "EVENTS", "THEMES", "CanvasThemes", "Themes", "get_settings", "is_dark"]
