from __future__ import annotations

from qtpy.QtWidgets import QLabel


class QtLargeLabel(QLabel):
    """Large label."""


class QtWelcomeLabel(QLabel):
    """Labels used for main message in welcome page."""


class QtShortcut(QLabel):
    """Labels used for displaying shortcut information in welcome page."""


class QtShortcutLabel(QLabel):
    """Labels used for displaying shortcut information in welcome page."""
